"""
Unit tests for user_management management commands.
"""


import itertools

import ddt
import pytest
from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import Group
from django.core.management import CommandError, call_command
from django.test import TestCase

from edx_django_utils.user import generate_password

TEST_EMAIL = 'test@example.com'
TEST_USERNAME = 'test-user'
User = get_user_model()


@ddt.ddt
class TestManageUserCommand(TestCase):
    """
    Tests the `manage_user` command.
    """

    def test_user(self):
        """
        Ensures that users are created if they don't exist and reused if they do.
        """
        assert not list(User.objects.all())
        call_command('manage_user', TEST_USERNAME, TEST_EMAIL)
        user = User.objects.get(username=TEST_USERNAME)
        assert user.username == TEST_USERNAME
        assert user.email == TEST_EMAIL

        # check idempotency
        call_command('manage_user', TEST_USERNAME, TEST_EMAIL)
        assert [(TEST_USERNAME, TEST_EMAIL)] == [(u.username, u.email) for u in User.objects.all()]

    def test_remove(self):
        """
        Ensures that users are removed if they exist and exit cleanly otherwise.
        """
        User.objects.create(username=TEST_USERNAME, email=TEST_EMAIL)
        assert [(TEST_USERNAME, TEST_EMAIL)] == [(u.username, u.email) for u in User.objects.all()]
        call_command('manage_user', TEST_USERNAME, TEST_EMAIL, '--remove')
        assert not list(User.objects.all())

        # check idempotency
        call_command('manage_user', TEST_USERNAME, TEST_EMAIL, '--remove')
        assert not list(User.objects.all())

    def test_unusable_password(self):
        """
        Ensure that a user's password is set to an unusable_password.
        """
        user = User.objects.create(username=TEST_USERNAME, email=TEST_EMAIL)
        assert [(TEST_USERNAME, TEST_EMAIL)] == [(u.username, u.email) for u in User.objects.all()]
        user.set_password(generate_password())
        user.save()

        # Run once without passing --unusable-password and make sure the password is usable
        call_command('manage_user', TEST_USERNAME, TEST_EMAIL)
        user = User.objects.get(username=TEST_USERNAME, email=TEST_EMAIL)
        assert user.has_usable_password()

        # Make sure the user now has an unusable_password
        call_command('manage_user', TEST_USERNAME, TEST_EMAIL, '--unusable-password')
        user = User.objects.get(username=TEST_USERNAME, email=TEST_EMAIL)
        assert not user.has_usable_password()

        # check idempotency
        call_command('manage_user', TEST_USERNAME, TEST_EMAIL, '--unusable-password')
        assert not user.has_usable_password()

    def test_initial_password_hash(self):
        """
        Ensure that a user's password hash is set correctly when the user is created,
        and that it isn't touched for existing users.
        """
        initial_hash = make_password('hunter2')

        # Make sure the command aborts if the provided hash isn't a valid Django password hash
        with pytest.raises(CommandError) as exc_context:
            call_command('manage_user', TEST_USERNAME, TEST_EMAIL, '--initial-password-hash', 'invalid_hash')
        assert 'password hash' in str(exc_context.value).lower()

        # Make sure the hash gets set correctly for a new user
        call_command('manage_user', TEST_USERNAME, TEST_EMAIL, '--initial-password-hash', initial_hash)
        user = User.objects.get(username=TEST_USERNAME)
        assert user.password == initial_hash

        # Change the password
        new_hash = make_password('correct horse battery staple')
        user.password = new_hash
        user.save()

        # Verify that calling manage_user again leaves the password untouched
        call_command('manage_user', TEST_USERNAME, TEST_EMAIL, '--initial-password-hash', initial_hash)
        user = User.objects.get(username=TEST_USERNAME)
        assert user.password == new_hash

    def test_wrong_email(self):
        """
        Ensure that the operation is aborted if the username matches an
        existing user account but the supplied email doesn't match.
        """
        User.objects.create(username=TEST_USERNAME, email=TEST_EMAIL)
        with pytest.raises(CommandError) as exc_context:
            call_command('manage_user', TEST_USERNAME, 'other@example.com')
        assert 'do not match' in str(exc_context.value).lower()
        assert [(TEST_USERNAME, TEST_EMAIL)] == [(u.username, u.email) for u in User.objects.all()]

        # check that removal uses the same check
        with pytest.raises(CommandError) as exc_context:
            call_command('manage_user', TEST_USERNAME, 'other@example.com', '--remove')
        assert 'do not match' in str(exc_context.value).lower()
        assert [(TEST_USERNAME, TEST_EMAIL)] == [(u.username, u.email) for u in User.objects.all()]

    def test_same_email_varied_case(self):
        """
        Ensure that the operation continues if the username matches an
        existing user account and the supplied email differs only in cases.
        """
        User.objects.create(username=TEST_USERNAME, email=TEST_EMAIL.upper())
        call_command('manage_user', TEST_USERNAME, TEST_EMAIL.lower())
        user = User.objects.get(username=TEST_USERNAME)
        assert user.email == TEST_EMAIL.upper()

    @ddt.data(*itertools.product([(True, True), (True, False), (False, True), (False, False)], repeat=2))
    @ddt.unpack
    def test_bits(self, initial_bits, expected_bits):
        """
        Ensure that the 'staff' and 'superuser' bits are set according to the
        presence / absence of the associated command options, regardless of
        any previous state.
        """
        initial_staff, initial_super = initial_bits
        User.objects.create(
            username=TEST_USERNAME,
            email=TEST_EMAIL,
            is_staff=initial_staff,
            is_superuser=initial_super
        )

        expected_staff, expected_super = expected_bits
        args = [opt for bit, opt in ((expected_staff, '--staff'), (expected_super, '--superuser')) if bit]
        call_command('manage_user', TEST_USERNAME, TEST_EMAIL, *args)
        user = User.objects.all().first()
        assert user.is_staff == expected_staff
        assert user.is_superuser == expected_super

    @ddt.data(*itertools.product(('', 'a', 'ab', 'abc'), repeat=2))
    @ddt.unpack
    def test_groups(self, initial_groups, expected_groups):
        """
        Ensures groups assignments are created and deleted idempotently.
        """
        groups = {}
        for group_name in 'abc':
            groups[group_name] = Group.objects.create(name=group_name)

        user = User.objects.create(username=TEST_USERNAME, email=TEST_EMAIL)
        for group_name in initial_groups:
            user.groups.add(groups[group_name])

        call_command('manage_user', TEST_USERNAME, TEST_EMAIL, '-g', *expected_groups)
        actual_groups = [group.name for group in user.groups.all()]
        assert actual_groups == list(expected_groups)

    def test_nonexistent_group(self):
        """
        Ensures the command does not fail if specified groups cannot be found.
        """
        user = User.objects.create(username=TEST_USERNAME, email=TEST_EMAIL)
        groups = {}
        for group_name in 'abc':
            groups[group_name] = Group.objects.create(name=group_name)
            user.groups.add(groups[group_name])

        call_command('manage_user', TEST_USERNAME, TEST_EMAIL, '-g', 'b', 'c', 'd')
        actual_groups = [group.name for group in user.groups.all()]
        assert actual_groups == ['b', 'c']



"""
Functions that deal with settings related to plugins

Please remember to expose any new public methods in the `__init__.py` file.
"""
from logging import getLogger

from . import constants, registry, utils

log = getLogger(__name__)


def add_plugins(settings_path, project_type, settings_type):
    """
    Updates the module at the given ``settings_path`` with all Plugin
    Settings appropriate for the given project_type and settings_type.
    """
    settings_module = utils.import_module(settings_path)
    for plugin_settings in _iter_plugins(project_type, settings_type):
        settings_func = getattr(
            plugin_settings, constants.PLUGIN_APP_SETTINGS_FUNC_NAME
        )
        settings_func(settings_module)


def _iter_plugins(project_type, settings_type):
    """
    Yields Plugin Settings modules that are registered for the given
    project_type and settings_type.
    """
    for app_config in registry.get_plugin_app_configs(project_type):
        settings_config = _get_config(app_config, project_type, settings_type)
        if settings_config is None:
            log.debug(
                "Plugin Apps [Settings]: Did NOT find %s for %s and %s",
                app_config.name,
                project_type,
                settings_type,
            )
            continue

        plugin_settings_path = utils.get_module_path(
            app_config, settings_config, constants.PluginSettings
        )

        log.debug(
            "Plugin Apps [Settings]: Found %s for %s and %s",
            app_config.name,
            project_type,
            settings_type,
        )
        yield utils.import_module(plugin_settings_path)


def _get_config(app_config, project_type, settings_type):
    plugin_config = getattr(app_config, constants.PLUGIN_APP_CLASS_ATTRIBUTE_NAME, {})
    settings_config = plugin_config.get(constants.PluginSettings.CONFIG, {})
    project_type_settings = settings_config.get(project_type, {})
    return project_type_settings.get(settings_type)



"""
Mock views with a different module to enable testing of mapping
code_owner to modules. Trying to mock __module__ on a view was
getting too complex.
"""
from django.views.generic import View


class MockViewTest(View):
    """
    Mock view for use in testing.
    """



"""
Methods to get plugin apps

Please remember to expose any new public methods in the `__init__.py` file.
"""
from logging import getLogger

from . import constants, registry

log = getLogger(__name__)


def get_plugin_apps(project_type):
    """
    Returns a list of all registered Plugin Apps, expected to be added to
    the INSTALLED_APPS list for the given project_type.
    """
    plugin_apps = [
        "{module_name}.{class_name}".format(
            module_name=app_config.__module__, class_name=app_config.__name__,
        )
        for app_config in registry.get_plugin_app_configs(project_type)
        if getattr(app_config, constants.PLUGIN_APP_CLASS_ATTRIBUTE_NAME, None)
        is not None
    ]
    log.debug("Plugin Apps: Found %s", plugin_apps)
    return plugin_apps



"""
Urls utility functions

Please remember to expose any new public methods in the `__init__.py` file.
"""
from logging import getLogger

from django.urls import include, re_path

from . import constants, registry, utils

log = getLogger(__name__)


def _get_url(url_module_path, url_config):
    """
    function constructs the appropriate URL
    """
    namespace = url_config[constants.PluginURLs.NAMESPACE]
    app_name = url_config.get(constants.PluginURLs.APP_NAME)
    regex = url_config.get(constants.PluginURLs.REGEX, r"")

    if namespace:
        return re_path(regex, include((url_module_path, app_name), namespace=namespace))
    else:
        return re_path(regex, include(url_module_path))


def get_plugin_url_patterns(project_type):
    """
    Returns a list of all registered Plugin URLs, expected to be added to
    the URL patterns for the given project_type.
    """
    return [
        _get_url(url_module_path, url_config)
        for url_module_path, url_config in _iter_plugins(project_type)
    ]


def _iter_plugins(project_type):
    """
    Yields the module path and configuration for Plugin URLs registered for
    the given project_type.
    """
    for app_config in registry.get_plugin_app_configs(project_type):
        url_config = _get_config(app_config, project_type)
        if url_config is None:
            log.debug(
                "Plugin Apps [URLs]: Did NOT find %s for %s",
                app_config.name,
                project_type,
            )
            continue

        urls_module_path = utils.get_module_path(
            app_config, url_config, constants.PluginURLs
        )
        url_config[constants.PluginURLs.NAMESPACE] = url_config.get(
            constants.PluginURLs.NAMESPACE, app_config.name
        )
        url_config[constants.PluginURLs.APP_NAME] = app_config.name
        log.debug(
            "Plugin Apps [URLs]: Found %s with namespace %s for %s",
            app_config.name,
            url_config[constants.PluginURLs.NAMESPACE],
            project_type,
        )
        yield urls_module_path, url_config


def _get_config(app_config, project_type):
    plugin_config = getattr(app_config, constants.PLUGIN_APP_CLASS_ATTRIBUTE_NAME, {})
    url_config = plugin_config.get(constants.PluginURLs.CONFIG, {})
    return url_config.get(project_type)



"""
Allows overriding existing functions and methods with alternative implementations.
"""

import functools
from importlib import import_module

from django.conf import settings


def pluggable_override(override):
    """
    This decorator allows overriding any function or method by pointing to an alternative implementation
    with `override` param.
    :param override: path to the alternative function

    Example usage:

    1. Add this decorator to an existing function `OVERRIDE_TRANSFORM` is the variable name in settings that can be
       used for overriding this function. Remember to add the `OVERRIDE_` prefix to the name to have the consistent
       namespace for the overrides.
        >>> @pluggable_override('OVERRIDE_TRANSFORM')
        ... def transform(value):
        ...     return value + 10

    2. Prepare an alternative implementation. It will have the same set of arguments as the original function, with the
       `prev_fn` added at the beginning.
        >>> def decrement(prev_fn, value):
        ...     if value >= 10:
        ...         return value - 1  # Return the decremented value.
        ...     else:
        ...         return prev_fn(value) - 1  # Call the original `transform` method before decrementing and returning.

    3. Specify the path in settings (e.g. in `envs/private.py`):
        >>> OVERRIDE_TRANSFORM = 'transform_plugin.decrement'

        You can also chain overrides:
        >>> OVERRIDE_TRANSFORM = [
        ...     'transform_plugin.decrement',
        ...     'transform_plugin.increment',
        ... ]

    Another example:

    1. We want to limit access to a Django view (e.g. `common.djangoapps.student.views.dashboard.student_dashboard`)
       to allow only logged in users. To do this add `OVERRIDE_DASHBOARD` to the original function:
       >>> @pluggable_override('OVERRIDE_DASHBOARD')
       ... def student_dashboard(request):
       ...     ...  # The rest of the implementation is not relevant in this case.

   2. Prepare an alternative implementation (e.g. in `envs/private.py` to make this example simpler):
      >>> from django.contrib.auth.decorators import login_required
      ...
      ... def dashboard(prev_fn, request):
      ...     return login_required(prev_fn)(request)
      ...
      ... OVERRIDE_DASHBOARD = 'lms.envs.private.dashboard'
    """
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            prev_fn = f

            override_functions = getattr(settings, override, ())

            if isinstance(override_functions, str):
                override_functions = [override_functions]

            for impl in override_functions:
                module, function = impl.rsplit('.', 1)
                mod = import_module(module)
                func = getattr(mod, function)

                prev_fn = functools.partial(func, prev_fn)
            # Call the last specified function. It can call the previous one, which can call the previous one, etc.
            # (until it reaches the base implementation). It can also return without calling `prev_fn`.
            return prev_fn(*args, **kwargs)
        return wrapper
    return decorator



"""
Mixins for use in Django Admin classes.
"""


class ReadOnlyAdminMixin:
    """
    Disables all editing capabilities for the admin's model.
    An example usage: a Django proxy model which provides data to perform some other action.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.list_display_links = None
        self.readonly_fields = [f.name for f in self.model._meta.get_fields()]

    def get_actions(self, request):
        actions = super().get_actions(request)
        if 'delete_selected' in actions:
            del actions["delete_selected"]  # pragma: no cover
        return actions

    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):  # pylint: disable=unused-argument
        return False

    def save_model(self, request, obj, form, change):
        pass  # pragma: no cover

    def delete_model(self, request, obj):
        pass  # pragma: no cover

    def save_related(self, request, form, formsets, change):
        pass  # pragma: no cover



"""
Tests of edx_django_utils.db.read_replica.
"""

from importlib import reload

from django.conf import settings
from django.contrib import auth
from django.test import TestCase, override_settings

from .. import read_replica

WRITER_ONLY_DATABASES = settings.DATABASES.copy()
del WRITER_ONLY_DATABASES["read_replica"]

READ_REPLICA_DATABASES = settings.DATABASES
DATABASE_ROUTERS = settings.DATABASE_ROUTERS + [
    "edx_django_utils.db.read_replica.ReadReplicaRouter"
]
User = auth.get_user_model()


class TestReadReplica(TestCase):
    """
    Tests of edx_django_utils.db.read_replica.
    """

    databases = ["default", "read_replica"]

    def setUp(self):
        super().setUp()
        self.addCleanup(reload, read_replica)

    def test_read_inside_write_error(self):
        with self.assertRaises(AssertionError):
            with read_replica.write_queries():
                with read_replica.read_queries_only():
                    pass  # pragma: no cover

    def test_write_inside_read_error(self):
        with self.assertRaises(AssertionError):
            with read_replica.read_queries_only():
                with read_replica.write_queries():
                    pass  # pragma: no cover

    def test_write_inside_write_ok(self):
        with read_replica.write_queries():
            with read_replica.write_queries():
                pass

    def test_read_inside_read_ok(self):
        with read_replica.read_queries_only():
            with read_replica.read_queries_only():
                pass

    def test_read_replica_name_from_default(self):
        reload(read_replica)
        assert read_replica.READ_REPLICA_NAME == "read_replica"

    @override_settings(EDX_READ_REPLICA_DB_NAME="new_read_replica")
    def test_read_replica_name_from_settings(self):
        reload(read_replica)
        assert read_replica.READ_REPLICA_NAME == "new_read_replica"

    @override_settings(DATABASES=WRITER_ONLY_DATABASES)
    def test_writer_name_from_default(self):
        reload(read_replica)
        assert read_replica.WRITER_NAME == "default"

    @override_settings(EDX_WRITER_DB_NAME="new_default")
    def test_writer_name_from_settings(self):
        reload(read_replica)
        assert read_replica.WRITER_NAME == "new_default"

    @override_settings(DATABASES=WRITER_ONLY_DATABASES)
    def test_read_replica_fallback(self):
        reload(read_replica)
        assert read_replica.READ_REPLICA_OR_DEFAULT == read_replica.WRITER_NAME

    @override_settings(DATABASES=READ_REPLICA_DATABASES)
    def test_read_replica_exists(self):
        reload(read_replica)
        assert read_replica.READ_REPLICA_OR_DEFAULT == read_replica.READ_REPLICA_NAME

    @override_settings(
        DATABASES=READ_REPLICA_DATABASES, DATABASE_ROUTERS=DATABASE_ROUTERS
    )
    def test_read_only_queries_from_database(self):
        reload(read_replica)
        User(username="test_user").save(using=read_replica.WRITER_NAME)

        assert User.objects.all().count() == 1
        with read_replica.read_queries_only():
            assert User.objects.all().count() == 0

    @override_settings(
        DATABASES=READ_REPLICA_DATABASES, DATABASE_ROUTERS=DATABASE_ROUTERS
    )
    def test_router_writes(self):
        reload(read_replica)
        User(username="test_user").save()

        assert User.objects.all().count() == 1
        with read_replica.read_queries_only():
            assert User.objects.all().count() == 0



"""
Tests for IP determination.

Fake data used in these tests, for consistency:

- 1.2.3.4 -- a "real" client IP, e.g. the IP of a laptop or phone.
- 127.0.0.2 -- a local reverse proxy (e.g. nginx or caddy)
- 10.0.3.0 -- our load balancer
- 5.5.5.5 -- our CDN
- 6.6.6.6 -- a malicious CDN configuration
- 7.8.9.0 -- something beyond the real client in the IP chain, probably a spoofed header

...as well as IPv6 versions of these, e.g. 1:2:3:4:: and ::1.

XXXXXXXXX is used as a standin for anything unparseable (some kind of garbage).
"""

import warnings
from contextlib import contextmanager

import ddt
from django.test import TestCase
from django.test.client import RequestFactory
from django.test.utils import override_settings

from edx_django_utils.ip.internal import ip


@contextmanager
def warning_messages():
    """
    Context manager which produces a list of warning messages as the context
    value (only populated after block ends).
    """
    with warnings.catch_warnings(record=True) as caught_warnings:
        warnings.simplefilter('always')
        messages = []
        yield messages
        # Converted to message strings for easier debugging
        messages.extend(str(w.message) for w in caught_warnings)


@ddt.ddt
class TestClientIP(TestCase):
    """Tests for get_client_ip and helpers."""

    def setUp(self):
        super().setUp()
        self.request = RequestFactory().get('/somewhere')

    @ddt.unpack
    @ddt.data(
        ({}, 'Some-Thing', []),
        ({'HTTP_SOME_THING': 'stuff'}, 'Some-Thing', ['stuff']),
        ({'HTTP_SOME_THING': 'stuff'}, 'Some-Thing', ['stuff']),
        ({'HTTP_SOME_THING': '   so,much , stuff '}, 'Some-Thing', ['so', 'much', 'stuff']),
    )
    def test_get_meta_ip_strs(self, add_meta, header_name, expected):
        self.request.META.update(add_meta)
        assert ip._get_meta_ip_strs(self.request, header_name) == expected  # pylint: disable=protected-access

    @ddt.unpack
    @ddt.data(
        # Form the IP chain and parse it (notice the IPv6 is canonicalized)
        (
            {'HTTP_X_FORWARDED_FOR': '7.8.9.0, 1.2.3.4, 10.0.3.0', 'REMOTE_ADDR': '0:0:0:0::1'},
            ['7.8.9.0', '1.2.3.4', '10.0.3.0', '::1'],
        ),

        # XFF is not required
        ({'REMOTE_ADDR': '127.0.0.2'}, ['127.0.0.2']),

        # Strips off junk and anything to the left of it
        (
            {
                'HTTP_X_FORWARDED_FOR': '6.6.6.6, XXXXXXXXX, 7.8.9.0, XXXXXXXXX, 1.2.3.4',
                'REMOTE_ADDR': '10.0.3.0'
            },
            ['1.2.3.4', '10.0.3.0'],
        ),
    )
    def test_get_usable_ip_chain(self, request_meta, expected_strs):
        self.request.META = request_meta
        actual_ips = ip._get_usable_ip_chain(self.request)  # pylint: disable=protected-access
        assert [str(ip) for ip in actual_ips] == expected_strs

    @ddt.unpack
    @ddt.data(
        # Empty returns empty
        ([], lambda x: False, []),
        ([], lambda x: True, []),

        # Can return whole list or remove all of it
        ([1, 2, 3, 4], lambda x: False, [1, 2, 3, 4]),
        ([1, 2, 3, 4], lambda x: True, []),

        # Walks from right, stops on first false, keeps that element
        ([2, 0, 1, 0, 3, 5], lambda x: x != 0, [2, 0, 1, 0]),
    )
    def test_remove_tail(self, elements, f_discard, expected):
        assert ip._remove_tail(elements, f_discard) == expected  # pylint: disable=protected-access

    @ddt.unpack
    @ddt.data(
        # Walk from right, dropping private-range IPs
        ('7:8:9:0::, 1.2.3.4, 10.0.3.0', '::1', ['7:8:9::', '1.2.3.4']),
        ('7.8.9.0', '1.2.3.4', ['7.8.9.0', '1.2.3.4']),

        # Publicly exposed server (no XFF added by proxies)
        (None, '1.2.3.4', ['1.2.3.4']),

        # If XFF is missing, just accept a private IP (maybe an inter-service call
        # from another server in the same datacenter)
        (None, '127.0.0.2', ['127.0.0.2']),

        # If we reach a public IP, don't worry about junk farther on
        ('XXXXXXXXX, 1:2:3:4::', '10.0.0.1', ['1:2:3:4::']),

        # If we find junk or we run out of IPs before finding a public
        # one, the best we can do is a private IP. (Should never
        # happen for a public IP, but here for completeness.)
        ('7.8.9.0, XXXXXXXXX, 10.0.3.0', '127.0.0.2', ['10.0.3.0']),
        (None, '::1', ['::1']),

        # Nothing usable (again, should never happen)
        (None, '', []),
        (None, 'XXXXXXXXX', []),
        ('1.2.3.4', 'XXXXXXXXX', []),
    )
    def test_get_client_ips_via_xff(self, xff, remote_addr, expected_strs):
        request_meta = {'REMOTE_ADDR': remote_addr, 'HTTP_X_FORWARDED_FOR': xff}
        request_meta = {k: v for k, v in request_meta.items() if v is not None}
        self.request.META = request_meta

        assert [str(ip) for ip in ip._get_client_ips_via_xff(self.request)] == expected_strs  # pylint: disable=protected-access

    @ddt.unpack
    @ddt.data(
        # Happy path
        ('Some-Thing', 0, {'HTTP_SOME_THING': '1.2.3.4'}, '1.2.3.4', None),
        ('some-thing', -1, {'HTTP_SOME_THING': '1:2:3:4::, 0:0::1'}, '::1', None),

        # Warning: Header present, index out of range
        ('Some-Thing', 1, {'HTTP_SOME_THING': '1.2.3.4'}, None, "out of range"),
        ('Some-Thing', -2, {'HTTP_SOME_THING': '1.2.3.4'}, None, "out of range"),

        # Warning: Header missing entirely
        ('Some-Thing', 0, {}, None, "missing"),

        # Warning: Bad IP address
        ('Some-Thing', 0, {'HTTP_SOME_THING': 'XXXXXXXXX'}, None, "invalid IP"),
    )
    def test_get_trusted_header_ip(self, header_name, index, add_meta, expected, warning_substr):
        self.request.META.update(add_meta)

        with warning_messages() as caught_warnings:
            actual = ip._get_trusted_header_ip(self.request, header_name, index)  # pylint: disable=protected-access

        if expected is None:
            assert actual is None
        else:
            assert str(actual) == expected  # Stringify again for comparison

        if warning_substr is None:
            assert len(caught_warnings) == 0
        else:
            assert len(caught_warnings) == 1
            assert warning_substr in caught_warnings[0]

    @ddt.unpack
    @ddt.data(
        # Most common case: One header in config, and header does exist
        (
            [{'name': 'CF-Connecting-IP', 'index': 0}],
            {
                'HTTP_X_FORWARDED_FOR': '1.2.3.4',
                'REMOTE_ADDR': '10.0.3.0',
                'HTTP_CF_CONNECTING_IP': '1.2.3.4',
            },
            ['1.2.3.4'],
            0,
        ),
        # More complicated version with intervening proxies and a spoofed IP
        (
            [{'name': 'CF-Connecting-IP', 'index': 0}],
            {
                'HTTP_X_FORWARDED_FOR': '7.8.9.0, 1.2.3.4, 5.5.5.5, 10.0.3.0',
                'REMOTE_ADDR': '127.0.0.2',
                'HTTP_CF_CONNECTING_IP': '  1.2.3.4  ',  # tests that whitespace is stripped
            },
            ['7.8.9.0', '1.2.3.4'],
            0,
        ),

        # Uses *rightmost* position of identified IP if multiple are present
        # (prevent spoofing when client calling through a trustworthy proxy)
        (
            [{'name': 'CF-Connecting-IP', 'index': 0}],
            {
                'HTTP_X_FORWARDED_FOR': '6.6.6.6, 1.2.3.4, 7.8.9.0, 1.2.3.4, 5.5.5.5',
                'REMOTE_ADDR': '10.0.3.0',
                'HTTP_CF_CONNECTING_IP': '1.2.3.4',
            },
            ['6.6.6.6', '1.2.3.4', '7.8.9.0', '1.2.3.4'],
            0,
        ),

        # No config? Empty list.
        ([], {'REMOTE_ADDR': '1.2.3.4'}, [], 0),
        (None, {'REMOTE_ADDR': '1.2.3.4'}, [], 0),

        # One lookup failure (a warning) before finding a usable header
        (
            [
                {'name': 'X-Real-IP', 'index': 0},
                {'name': 'CF-Connecting-IP', 'index': 0},
            ],
            {
                'HTTP_X_FORWARDED_FOR': '7.8.9.0, 1.2.3.4, 5.5.5.5',
                'REMOTE_ADDR': '10.0.3.0',
                'HTTP_CF_CONNECTING_IP': '1.2.3.4',
            },
            ['7.8.9.0', '1.2.3.4'],
            1,
        ),

        # Configured, but none of the headers are present
        (
            [
                {'name': 'CF-Connecting-IP', 'index': 0},
                {'name': 'X-Forwarded-For', 'index': -2},
            ],
            {'REMOTE_ADDR': '1.2.3.4'},
            [],
            2,
        ),

        # Can configure to use far end of XFF if needed for some reason
        (
            [
                {'name': 'X-Forwarded-For', 'index': 0},
            ],
            {
                'HTTP_X_FORWARDED_FOR': '1.2.3.4, 5.5.5.5',
                'REMOTE_ADDR': '10.0.3.0',
            },
            ['1.2.3.4'],
            0,
        ),

        # Header is present but doesn't match anything in the XFF
        (
            [
                {'name': 'X-Real-IP', 'index': 0},
            ],
            {
                'HTTP_X_FORWARDED_FOR': '7.8.9.0, 1.2.3.4, 5.5.5.5',
                'REMOTE_ADDR': '10.0.3.0',
                'HTTP_X_REAL_IP': '127.0.0.1',
            },
            [],
            1,
        ),
    )
    def test_get_client_ips_via_trusted_header(self, cnf_headers, add_meta, expected, warning_count):
        self.request.META.update(add_meta)
        if cnf_headers is None:
            overrides = {}
        else:
            overrides = {'CLOSEST_CLIENT_IP_FROM_HEADERS': cnf_headers}

        with override_settings(**overrides):
            with warning_messages() as caught_warnings:
                actual = ip._get_client_ips_via_trusted_header(self.request)  # pylint: disable=protected-access

        assert [str(ip) for ip in actual] == expected
        assert len(caught_warnings) == warning_count

    @ddt.unpack
    @ddt.data(
        # Using headers setting
        (
            [{'name': 'CF-Connecting-IP', 'index': 0}],
            {
                'HTTP_X_FORWARDED_FOR': '7.8.9.0, 1.2.3.4, 5.5.5.5, 10.0.3.0',
                'HTTP_CF_CONNECTING_IP': '1.2.3.4',
                'REMOTE_ADDR': '127.0.0.2',
            },
            ['7.8.9.0', '1.2.3.4'],
            0,
        ),

        # Fall back when all override headers are unusable, with warnings
        (
            [
                {'name': 'CF-Connecting-IP', 'index': 0},  # not actually passed in this request
                {'name': 'X-Real-IP', 'index': 2},  # index out of range, so unusable
            ],
            {
                'HTTP_X_FORWARDED_FOR': '7.8.9.0, 1.2.3.4, 10.0.3.0, 127.0.0.2',
                'HTTP_X_REAL_IP': '10.0.3.0',
                'REMOTE_ADDR': '127.0.0.2',
            },
            ['7.8.9.0', '1.2.3.4'],
            2,
        ),

        # By default, with the least possible information, do something useful.
        # (And no warnings when no override headers are specified.)
        (None, {'REMOTE_ADDR': '127.0.0.2'}, ['127.0.0.2'], 0),
    )
    def test_compute_client_ips(self, cnf_headers, add_meta, expected, expect_warnings):
        """
        Just a few tests to confirm that the correct branch is taken, basically.
        """
        if cnf_headers is None:
            overrides = {}
        else:
            overrides = {'CLOSEST_CLIENT_IP_FROM_HEADERS': cnf_headers}

        self.request.META.update(add_meta)

        with override_settings(**overrides):
            with warning_messages() as caught_warnings:
                actual = ip._compute_client_ips(self.request)  # pylint: disable=protected-access

        assert actual == expected
        assert len(caught_warnings) == expect_warnings

    def test_init_client_ips(self):
        """
        Test idempotence of init_client_ips.
        """
        self.request.META.update({
            'HTTP_X_FORWARDED_FOR': '1.2.3.4',
            'REMOTE_ADDR': '127.0.0.2',
        })

        # Not there to begin with
        assert 'CLIENT_IPS' not in self.request.META

        # Gets added by init call
        with warning_messages() as caught_warnings:
            ip.init_client_ips(self.request)

        assert self.request.META['CLIENT_IPS'] == ['1.2.3.4']
        assert len(caught_warnings) == 0

        # Second call warns but does not recalculate.
        # (This input wouldn't change in reality; just changing it here to illustrate the caching.)
        self.request.META['HTTP_X_FORWARDED_FOR'] = '7.8.9.0, 1.2.3.4'
        with warning_messages() as caught_warnings:
            ip.init_client_ips(self.request)

        assert self.request.META['CLIENT_IPS'] == ['1.2.3.4']  # unchanged
        assert len(caught_warnings) == 1

    def test_get_all_client_ips(self):
        """
        Test getter for all IPs.
        """
        self.request.META.update({
            'HTTP_X_FORWARDED_FOR': '7.8.9.0, 1.2.3.4',
            'REMOTE_ADDR': '127.0.0.2',
        })

        assert 'CLIENT_IPS' not in self.request.META
        assert ip.get_all_client_ips(self.request) == ['7.8.9.0', '1.2.3.4']
        assert 'CLIENT_IPS' in self.request.META

        # Just calling again for coverage (cached case)
        assert ip.get_all_client_ips(self.request) == ['7.8.9.0', '1.2.3.4']

    def test_get_safest_client_ip(self):
        """
        Test convenience wrapper for rightmost IP.
        """
        self.request.META.update({
            'HTTP_X_FORWARDED_FOR': '7.8.9.0',
            'REMOTE_ADDR': '1.2.3.4',
        })

        assert ip.get_safest_client_ip(self.request) == '1.2.3.4'



"""
Django-based logging filters
"""

from logging import Filter

from crum import get_current_request, get_current_user


class RemoteIpFilter(Filter):
    """
    A logging filter that adds the remote IP to the logging context
    """
    def filter(self, record):
        request = get_current_request()
        if request and 'REMOTE_ADDR' in request.META:
            record.remoteip = request.META['REMOTE_ADDR']
        else:
            record.remoteip = None
        return True


class UserIdFilter(Filter):
    """
    A logging filter that adds userid to the logging context
    """
    def filter(self, record):
        user = get_current_user()
        if user and user.pk:
            record.userid = user.pk
        else:
            record.userid = None
        return True



"""
Tools for making queries read from the read-replica database, rather than from the writer database.

The read-replica can be used for
    a. long-running queries that aren't time sensitive
    b. reads of rows that are frequently written, but where reads can be eventually consistent

Settings:
    EDX_READ_REPLICA_DB_NAME: The name of the read-replica in the DATABASES django setting.
        Defaults to "read_replica".
    EDX_WRITER_DB_NAME: The name of the writer in the DATABASES django setting.
        Defaults to "default".
"""
import threading
from contextlib import contextmanager

from django.conf import settings

READ_REPLICA_NAME = getattr(settings, "EDX_READ_REPLICA_DB_NAME", "read_replica")
WRITER_NAME = getattr(settings, "EDX_WRITER_DB_NAME", "default")

READ_REPLICA_OR_DEFAULT = (
    READ_REPLICA_NAME if READ_REPLICA_NAME in settings.DATABASES else WRITER_NAME
)


def use_read_replica_if_available(queryset):
    """
    If there is a database called 'read_replica',
    use that database for the queryset / manager.

    Example usage:
        queryset = use_read_replica_if_available(SomeModel.objects.filter(...))

    Arguments:
        queryset (QuerySet)

    Returns: QuerySet
    """
    return queryset.using(READ_REPLICA_OR_DEFAULT)


def read_replica_or_default():
    """
    If there is a database called READ_REPLICA_DB,
    return READ_REPLICA_DB, otherwise return WRITER_NAME.

    This function is similiar to `use_read_replica_if_available`,
    but is be more syntactically convenient for method call chaining.
    Also, it always falls back to WRITER_NAME,
    no matter what the queryset was using before.

    Example usage:
        queryset = SomeModel.objects.filter(...).using(read_replica_or_default())

    Returns: str
    """
    return READ_REPLICA_OR_DEFAULT


@contextmanager
def read_queries_only():
    """
    A context manager that sets all reads inside it to be from the read-replica.

    It is an error to call this from inside a write_queries context.

    The ReadReplicaRouter must be used for this decorator to affect queries.
    """
    old_db_name = _storage.db_name
    assert (
        old_db_name is None or old_db_name == READ_REPLICA_NAME
    ), "Can't use read_queries_only inside a write_queries contextmanager"
    _storage.db_name = READ_REPLICA_NAME
    try:
        yield
    finally:
        _storage.db_name = old_db_name


@contextmanager
def write_queries():
    """
    A context manager that sets all reads inside it to be from the writer.
    Use this to annotate code that has both reads and writes, where the writes depend
    on the values read. This will allow all of that code to exist within a transaction.

    Using this contextmanager will prevent any contained call from using `read_queries_only`
    in order to read from the read-replica.

    It is an error to call this from inside a read_queries_only context.

    The ReadReplicaRouter must be used for this decorator to affect queries.
    """
    old_db_name = _storage.db_name
    assert (
        old_db_name is None or old_db_name == WRITER_NAME
    ), "Can't use write_queries inside a read_only_queries contextmanager"

    _storage.db_name = WRITER_NAME
    try:
        yield
    finally:
        _storage.db_name = old_db_name


class _ReadReplicaRouterStorage(threading.local):
    def __init__(self):
        super().__init__()
        self.db_name = None


_storage = _ReadReplicaRouterStorage()


class ReadReplicaRouter:
    """
    A database router that by default, reads from the writer database,
    but can be overridden with a context manager to route all reads
    to the read-replica.

    See https://docs.djangoproject.com/en/2.2/topics/db/multi-db/#automatic-database-routing
    """

    def db_for_read(self, model, **hints):  # pylint: disable=unused-argument
        """
        Reads go the active reader name
        """
        return (
            _storage.db_name if _storage.db_name in settings.DATABASES else WRITER_NAME
        )

    def db_for_write(self, model, **hints):  # pylint: disable=unused-argument
        """
        Writes always go to the writer.
        """
        return WRITER_NAME

    def allow_relation(
        self, obj1, obj2, **hints
    ):  # pylint: disable=unused-argument, protected-access
        """
        Relations between objects are allowed if both objects are
        in either the read-replica or the writer.
        """
        db_list = (READ_REPLICA_NAME, WRITER_NAME)
        if obj1._state.db in db_list and obj2._state.db in db_list:
            return True
        return None

    def allow_migrate(
        self, db, app_label, model_name=None, **hints
    ):  # pylint: disable=unused-argument
        """
        All non-auth models end up in this pool.
        """
        return True



"""
This script will process logs generated from CookieMonitoringMiddleware.

Sample usage::

    python edx_django_utils/monitoring/scripts/process_cookie_monitoring_logs.py --csv_input large-cookie-logs.csv

Or for more details::

    python edx_django_utils/monitoring/scripts/process_cookie_monitoring_logs.py --help

"""
import csv
import logging
import re

import click
from dateutil import parser

logging.basicConfig(format="%(levelname)s:%(message)s", level=logging.INFO)


# Note: Not all Open edX deployments will be affected by the same third-party cookies,
#   but it is ok to have some of these cookies go unused.
PARAMETERIZED_COOKIES = [
    (re.compile(r"_gac_UA-(\d|-)+"), "_gac_UA-{id}"),
    (re.compile(r"_hjSession_\d+"), "_hjSession_{id}"),
    (re.compile(r"_hjSessionUser_\d+"), "_hjSessionUser_{id}"),
    (re.compile(r"ab\.storage\.deviceId\..*"), "ab.storage.deviceId.{id}"),
    (re.compile(r"ab\.storage\.sessionId\..*"), "ab.storage.deviceId.{id}"),
    (re.compile(r"ab\.storage\.userId\..*"), "ab.storage.userId.{id}"),
    (re.compile(r"AMCV_\w+%40AdobeOrg"), "AMCV_{id}@AdobeOrg"),
    (re.compile(r"amplitude_id_.*"), "amplitude_id_{id}"),
    (re.compile(r"mp_\w+_mixpanel"), "mp_{id}_mixpanel"),
]


@click.command()
@click.option(
    "--csv_input",
    help="File name of .csv file with Splunk logs for large cookie headers.",
    required=True
)
def main(csv_input):
    """
    Reads CSV of large cookie logs and processes and provides summary output.

    Expected CSV format (from Splunk export searching on "BEGIN-COOKIE-SIZES"):

        \b
        _raw,_time,index,
        ...

    """
    cookie_headers = _load_csv(csv_input)
    processed_cookie_headers = process_cookie_headers(cookie_headers)
    print_processed_cookies(processed_cookie_headers)


def _load_csv(csv_file):
    """
    Reads CSV of large cookie data and returns a dict of details.

    Arguments:
        csv_file (string): File name for the csv

    Returns a list of dicts containing parsed details for each cookie header log entry.

    """
    with open(csv_file) as file:
        csv_data = file.read()
    reader = csv.DictReader(csv_data.splitlines())

    # Regex to match against log messages like the following:
    #   BEGIN-COOKIE-SIZES(total=3773) user-info: 903, csrftoken: 64, ... END-COOKIE-SIZES
    # Use a non-greedy (*?) quantifier for the occasional logs that have >1 set of cookie data
    cookie_log_regex = re.compile(r"BEGIN-COOKIE-SIZES\(total=(?P<total>\d+)\)(?P<cookie_sizes>.*?)END-COOKIE-SIZES")
    # Regex to match against just a single size, like the following:
    #   csrftoken: 64
    cookie_size_regex = re.compile(r"(?P<name>.*): (?P<size>\d+)")
    cookie_headers = []
    for row in reader:
        cookie_header_sizes = {}

        raw_cookie_log = row.get("_raw")
        cookie_begin_count = raw_cookie_log.count("BEGIN-COOKIE-SIZES")
        if cookie_begin_count == 0:
            logging.info("No BEGIN-COOKIE-SIZES delimiter found. Skipping row.")
            continue
        matches = cookie_log_regex.findall(raw_cookie_log)
        if len(matches) == 0:
            logging.error("Malformed cookie entry. Skipping row.")
            continue
        for match in matches:
            cookie_header_size = int(match[0])
            if cookie_header_size == 0:
                continue

            cookie_sizes_str = match[1].strip()

            cookie_sizes = cookie_sizes_str.split(", ")
            for cookie_size in cookie_sizes:
                match = cookie_size_regex.search(cookie_size)
                if not match:
                    logging.error(f"Could not parse cookie size from: {cookie_size}")
                    continue
                cookie_header_sizes[match.group("name")] = int(match.group("size"))

                cookie_header_size_computed = max(
                    0, sum(len(name) + size + 3 for (name, size) in cookie_header_sizes.items()) - 2
                )

            cookie_headers.append({
                "datetime": parser.parse(row.get("_time")),
                "env": row.get("index"),
                "cookie_header_size": cookie_header_size,
                "cookie_header_size_computed": cookie_header_size_computed,
                "cookie_sizes": cookie_header_sizes,
            })

    return cookie_headers


def process_cookie_headers(cookie_headers):
    """
    Process the parsed cookie header log entries.

    Arguments:
        cookie_headers: a list of dicts containing parsed details.

    Returns a dict of processed cookies.
    """
    processed_cookies = {}
    for cookie_header in cookie_headers:
        for (name, size) in cookie_header["cookie_sizes"].items():

            # Replace parameterized cookies. For example:
            #   _hjSessionUser_111111 => _hjSessionUser_{id}
            for (regex, replacement_name) in PARAMETERIZED_COOKIES:
                if regex.fullmatch(name):
                    logging.debug(f"Replacing {name} with {replacement_name}.")
                    name = replacement_name
                    break

            processed_cookie = processed_cookies.get(name, {})
            # compute the full size each cookie takes up in the cookie header, including name and delimiters
            full_size = len(name) + size + 3
            set_max_attribute(processed_cookie, "max_full_size", full_size)
            set_min_attribute(processed_cookie, "min_full_size", full_size)

            set_max_attribute(processed_cookie, "max_size", size)
            set_min_attribute(processed_cookie, "min_size", size)

            processed_cookie["count"] = processed_cookie.get("count", 0) + 1

            # Note: The following details relate to the header, and not to the specific cookie.
            #   This may give a quick view of cookies associated with the largest header, or the
            #   header with the most cookies.

            set_max_attribute(processed_cookie, "last_seen", cookie_header["datetime"])
            set_min_attribute(processed_cookie, "first_seen", cookie_header["datetime"])

            set_max_attribute(processed_cookie, "max_cookie_count", len(cookie_header["cookie_sizes"]))
            set_min_attribute(processed_cookie, "min_cookie_count", len(cookie_header["cookie_sizes"]))
            set_max_attribute(processed_cookie, "max_header_size", cookie_header["cookie_header_size"])
            set_min_attribute(processed_cookie, "min_header_size", cookie_header["cookie_header_size"])
            # Note: skipping cookie_header_size_calculated unless we see a need for it.

            processed_cookie["envs"] = processed_cookie.get("envs", set())
            processed_cookie["envs"].add(cookie_header["env"])

            processed_cookies[name] = processed_cookie

    return processed_cookies


def set_min_attribute(processed_cookie, key, value):
    """
    Sets processed_cookie[key] to the smaller of value and its current value.
    """
    processed_cookie[key] = min(value, processed_cookie.get(key, value))


def set_max_attribute(processed_cookie, key, value):
    """
    Sets processed_cookie[key] to the larger of value and its current value.
    """
    processed_cookie[key] = max(value, processed_cookie.get(key, value))


def print_processed_cookies(processed_cookies):
    """
    Output processed cookie information.
    """
    sorted_cookie_items = sorted(processed_cookies.items(), key=lambda x: x[1]["max_full_size"], reverse=True)
    print("name,max_full_size,min_full_size,max_size,min_size,count,"
          "last_seen,first_seen,max_cookie_count,min_cookie_count,max_header_size,min_header_size,envs")
    for (name, data) in sorted_cookie_items:
        print(f'{name},{data["max_full_size"]},{data["min_full_size"]},'
              f'{data["max_size"]},{data["min_size"]},{data["count"]},'
              f'{data["last_seen"]},{data["first_seen"]},'
              f'{data["max_cookie_count"]},{data["min_cookie_count"]},'
              f'{data["max_header_size"]},{data["min_header_size"]},'
              f'"{",".join(sorted(data["envs"]))}"')


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""
Implementation of ``edx_django_utils.ip`` -- contains non-public
functions, and is subject to breaking changes without warning.

Import from ``edx_django_utils.ip`` instead of this one.
See that module's docstring for additional background.
"""

import ipaddress
import warnings

from django.conf import settings


def _get_meta_ip_strs(request, header_name):
    """
    Get a list of IPs from a header in the given request.

    Return the list of IPs the request is carrying on this header, which is
    expected to be comma-delimited if it contains more than one. Response
    may be an empty list for missing or empty header. List items may not be
    valid IPs.
    """
    header_value = request.headers.get(header_name, '').strip()

    if header_value:
        return [s.strip() for s in header_value.split(',')]
    else:
        return []


def get_raw_ip_chain(request):
    """
    Retrieve the full IP chain from this request, as list of raw strings.

    This is uninterpreted and unparsed, except for splitting on commas and
    removing extraneous whitespace.
    """
    return _get_meta_ip_strs(request, 'X-Forwarded-For') + [request.META['REMOTE_ADDR']]


def _get_usable_ip_chain(request):
    """
    Retrieve the full IP chain from this request, as parsed addresses.

    The IP chain is the X-Forwarded-For header, followed by the REMOTE_ADDR.
    This list is then narrowed to the largest suffix that can be parsed as
    IP addresses.
    """
    parsed = []
    for ip_str in reversed(get_raw_ip_chain(request)):
        try:
            parsed.append(ipaddress.ip_address(ip_str))
        except ValueError:
            break
    return list(reversed(parsed))


def _remove_tail(elements, f_discard):
    """
    Remove items from the tail of the given list until f_discard returns false.

    - elements is a list
    - f_discard is a function that accepts an item from the list and returns
      true if it should be discarded from the tail

    Returns a new list that is a possibly-empty prefix of the input list.

    (This is basically itertools.dropwhile on a reversed list.)
    """
    prefix = elements[:]
    while prefix and f_discard(prefix[-1]):
        prefix.pop()
    return prefix


def _get_client_ips_via_xff(request):
    """
    Get the external chain of the request by discarding private IPs.

    This is a strategy used by ``get_all_client_ips`` and should not be used
    directly.

    Returns a list of *parsed* IP addresses, one of:

    - A list ending in a publicly routable IP
    - A list with a single, private-range IP
    - An empty list, if REMOTE_ADDR was unparseable as an IP address. This
      would be very unusual but could possibly happen if a local reverse proxy
      used a domain socket rather than a TCP connection.
    """
    ip_chain = _get_usable_ip_chain(request)
    external_chain = _remove_tail(ip_chain, lambda ip: not ip.is_global)

    # If the external_chain is in fact all private, everything will have been
    # removed. In that case, just return the leftmost IP it would have
    # considered, even though it must be a private IP.
    return external_chain or ip_chain[:1]


# .. setting_name: CLOSEST_CLIENT_IP_FROM_HEADERS
# .. setting_default: []
# .. setting_description: A list of header/index pairs to use for determining the IP in the
#   IP chain that is just outside of this deployment's infrastructure boundary -- that is,
#   the rightmost address in the IP chain that is *not* owned by the deployment. (See module
#   docstring for background and definitions, as well as guidance on configuration.)
#       Each list entry is a dict containing a header name and an index into that header. This will
#   control how the client's IP addresses are determined for attribution, tracking, rate-limiting,
#   or other general-purpose needs.
#       The named header must contain a list of IP addresses separated by commas, with whitespace
#   tolerated around each address. The index is used for a Python list lookup, e.g. 0 is the first
#   element and -2 is the second from the end.
#       Header/index pairs will be tried in turn until the first one that yields a usable IP, which
#   will then be used to determine the end of the external chain.
#       If the setting is an empty list, or if none of the entries yields a usable IP (header is
#   missing, index out of range, IP not in IP chain), then a fallback strategy will be used
#   instead: Private-range IPs will be discarded from the right of the IP chain until a public
#   IP is found, or the chain shrinks to one IP. This entry will then be considered the rightmost
#   end of the external chain.
#       Migrations from one network configuration to another may be accomplished by first adding the
#   new header to the list, making the networking change, and then removing the old one.
# .. setting_warnings: Changes to the networking configuration that are not coordinated with
#   this setting may allow callers to spoof their IP address.


def _get_trusted_header_ip(request, header_name, index):
    """
    Read a parsed IP address from a header at the specified position.

    Helper function for ``_get_client_ips_via_trusted_header``.

    Returns None if header is missing, index is out of range, or the located
    entry can't be parsed as an IP address.
    """
    ip_strs = _get_meta_ip_strs(request, header_name)

    if not ip_strs:
        warnings.warn(f"Configured IP address header was missing: {header_name!r}", UserWarning)
        return None

    try:
        trusted_ip_str = ip_strs[index]
    except IndexError:
        warnings.warn(
            "Configured index into IP address header is out of range: "
            f"{header_name!r}:{index!r} "
            f"(actual length {len(ip_strs)})",
            UserWarning
        )
        return None

    try:
        return ipaddress.ip_address(trusted_ip_str)
    except ValueError:
        warnings.warn(
            "Configured trusted IP address header contained invalid IP: "
            f"{header_name!r}:{index!r}",
            UserWarning
        )
        return None


def _get_client_ips_via_trusted_header(request) -> list:
    """
    Get the external chain by reading the trust boundary from a header.

    This is a strategy used by ``get_all_client_ips`` and should not be used
    directly. It does not implement any fallback in case of misconfiguration.

    Uses ``CLOSEST_CLIENT_IP_FROM_HEADERS`` to identify the IP just outside of
    the deployment's infrastructure boundary, and uses the rightmost position
    of this to determine where the external chain stops. See setting docs for
    more details.

    Returns one of the following:

    - A non-empty list of *parsed* IP addresses, where the rightmost IP is the
      same as the one identified in the trusted header.
    - Empty list if no headers configured or all headers are unusable.

    A configured header can be unusable if it's missing from the request, the
    index is out of range, the indicated entry in the header can't be parsed
    as an IP address, or the IP in the header can't be found in the IP chain.
    """
    header_entries = getattr(settings, 'CLOSEST_CLIENT_IP_FROM_HEADERS', [])

    full_chain = _get_usable_ip_chain(request)
    external_chain = []

    for entry in header_entries:
        header_name = entry['name']
        index = entry['index']
        if closest_client_ip := _get_trusted_header_ip(request, header_name, index):
            # The equality check in this predicate is why we use parsed IP
            # addresses -- ::1 should compare as equal to 0:0:0:0:0:0:0:1.
            external_chain = _remove_tail(full_chain, lambda ip: ip != closest_client_ip)  # pylint: disable=cell-var-from-loop
            if external_chain:
                break

            warnings.warn(
                f"Ignoring trusted header IP {header_name!r}:{index!r} "
                "because it was not found in the actual IP chain.",
                UserWarning
            )

    return external_chain


def _compute_client_ips(request):
    """
    Get the request's external chain, a non-empty list of IP address strings.

    Warning: should only be called once and cached by ``init_client_ips``.

    Prefer to use ``get_all_client_ips`` to retrieve the value stored on the
    request, unless you are sure that later middleware has not modified
    the REMOTE_ADDR in-place.

    This function will attempt several strategies to determine the external chain:

    - If ``CLOSEST_CLIENT_IP_FROM_HEADERS`` is configured and usable, it will be
      used to determine the rightmost end of the external chain (by reading a
      trusted HTTP header).
    - If that does not yield a result, fall back to assuming that the rightmost
      public IP address in the IP chain is the end of the external chain. (For an
      in-datacenter HTTP request, may instead yield a list with a private IP.)
    """
    # In practice the fallback to REMOTE_ADDR should never happen, since that
    # would require that value to be present and malformed but with no XFF
    # present.
    ips = _get_client_ips_via_trusted_header(request) \
        or _get_client_ips_via_xff(request) \
        or [request.META['REMOTE_ADDR']]

    return [str(ip) for ip in ips]


def init_client_ips(request):
    """
    Compute the request's external chain and store it in the request.

    This should be called early in the middleware stack in order to avoid
    being called after another middleware that overwrites ``REMOTE_ADDR``,
    which is a pattern some apps use.

    If called multiple times or if ``CLIENT_IPS`` is already present in
    ``request.META``, will just warn.
    """
    if 'CLIENT_IPS' in request.META:
        warnings.warn("init_client_ips refusing to overwrite existing CLIENT_IPS")
    else:
        request.META['CLIENT_IPS'] = _compute_client_ips(request)


def get_all_client_ips(request):
    """
    Get the request's external chain, a non-empty list of IP address strings.

    Most consumers of IP addresses should just use ``get_safest_client_ip``.

    Calls ``init_client_ips`` if needed.
    """
    if 'CLIENT_IPS' not in request.META:
        init_client_ips(request)

    return request.META['CLIENT_IPS']


def get_safest_client_ip(request):
    """
    Get the safest choice of client IP.

    Returns a single string containing the IP address that most likely
    represents the originator of the HTTP call, without compromising on
    safety.

    This is always the rightmost value in the external IP chain that
    is returned by ``get_all_client_ips``. See module docstring for
    more details.
    """
    return get_all_client_ips(request)[-1]



"""
Tests of edx_django_utils.db.queryset_utils.
"""
from ddt import data, ddt, unpack
from django.contrib import auth
from django.test import TestCase

from edx_django_utils.db.queryset_utils import chunked_queryset

User = auth.get_user_model()


@ddt
class TestQuerysetUtils(TestCase):
    """
    Tests of edx_django_utils.db.queryset_utils.
    """
    @unpack
    @data(
        (30, 10, [10, 10, 10]),
        (31, 10, [10, 10, 10, 1]),
        (10, 10, [10]),
        (7, 10, [7]),
        (0, 10, [0]),
    )
    def test_chunked_queryset(self, query_size, chunk_size, expected_batches):
        User.objects.all().delete()

        # create objects size of query_size
        for number in range(query_size):
            User.objects.create(username="username_{number}".format(number=number))

        queryset = User.objects.all()

        self.assertEqual(queryset.count(), query_size)
        for (batch_num, chunked_query) in enumerate(chunked_queryset(queryset, chunk_size)):
            self.assertEqual(chunked_query.count(), expected_batches[batch_num])

    def test_concurrent_update(self):
        """
        Test concurrent database modification wouldn't skip records.
        """
        User.objects.all().delete()

        # Create 14 objects.
        for number in range(14):
            User.objects.create(username="username_{number}".format(number=number))

        queryset = User.objects.all()

        # Now create chunks of size 10.
        chunked_query = chunked_queryset(queryset, chunk_size=10)

        # As there a total 14 objects and chunk size is 10, Assert first chunk should contain 10 objects.
        first_chunk = next(chunked_query)
        self.assertEqual(first_chunk.count(), 10)

        # Lets create a new object while iterating over the chunked_queryset.
        User.objects.create(username="one-more-user")

        # As now there are total 15 objects, the second chunk should contain 5 objects instead of 4.
        # that implies concurrent database modification won't skip records in this process.
        second_chunk = next(chunked_query)
        self.assertEqual(second_chunk.count(), 5)



"""
utils to help with imports

Please remember to expose any new public methods in the `__init__.py` file.
"""
from importlib import import_module as system_import_module

from django.utils.module_loading import import_string


def import_module(module_path):
    """
    Import and returns the module at the specific path.

    Args:
        module_path is the full path to the module, including the package name.
    """
    return system_import_module(module_path)


def get_module_path(app_config, plugin_config, plugin_cls):
    return "{package_path}.{module_path}".format(
        package_path=app_config.name,
        module_path=plugin_config.get(
            plugin_cls.RELATIVE_PATH, plugin_cls.DEFAULT_RELATIVE_PATH
        ),
    )


def import_attr(attr_path):
    """
    Import and returns a module's attribute at the specific path.

    Args:
        attr_path should be of the form:
            {full_module_path}.attr_name
    """
    return import_string(attr_path)


def import_attr_in_module(imported_module, attr_name):
    """
    Import and returns the attribute with name attr_name
    in the given module.
    """
    return getattr(imported_module, attr_name)



"""
Utils related to QuerySets.
"""


def chunked_queryset(queryset, chunk_size=2000):
    """
    Slice a queryset into chunks.

    The function slices a queryset into smaller QuerySets containing chunk_size objects and then yields them. It is
    used to avoid memory error when processing huge querysets, and also to avoid database errors due to the
    database pulling the whole table at once. Additionally, without using a chunked queryset, concurrent database
    modification while processing a large table might repeat or skip some entries.

    Warning: It throws away your sorting and sort queryset based on `pk`. Only recommended for large QuerySets where
    order does not matter.
    (e.g: Can be used in management commands to back-fill data based on Queryset having millions of objects.)

    Source: https://www.djangosnippets.org/snippets/10599/

    Example Usage:
        queryset = User.objects.all()
        for chunked_queryset in chunked_queryset(queryset):
            print(chunked_queryset.count())

    Argument:
        chunk_size (int): Size of desired batch.

    Return:
        QuerySet: Iterator with sliced Queryset.
    """
    start_pk = 0
    queryset = queryset.order_by('pk')

    while True:
        # No entry left
        if not queryset.filter(pk__gt=start_pk).exists():
            return

        try:
            # Fetch chunk_size entries if possible
            end_pk = queryset.filter(pk__gt=start_pk).values_list('pk', flat=True)[chunk_size - 1]

            # Fetch rest entries if less than chunk_size left
        except IndexError:
            end_pk = queryset.values_list('pk', flat=True).last()

        yield queryset.filter(pk__gt=start_pk).filter(pk__lte=end_pk)

        start_pk = end_pk



"""
Defines several custom monitoring helpers, some of which work with
the CachedCustomMonitoringMiddleware.

Usage:

    from edx_django_utils.monitoring import accumulate
    ...
    accumulate('xb_user_state.get_many.num_items', 4)

There is no need to do anything else.  The custom attributes are automatically
cleared before the next request.

We try to keep track of our custom monitoring at:
https://openedx.atlassian.net/wiki/spaces/PERF/pages/54362736/Custom+Attributes+in+New+Relic

At this time, the custom monitoring will only be reported to New Relic.

"""
from .middleware import CachedCustomMonitoringMiddleware

try:
    import newrelic.agent
except ImportError:  # pragma: no cover
    newrelic = None  # pylint: disable=invalid-name


def accumulate(name, value):
    """
    Accumulate monitoring custom attribute for the current request.

    The named attribute is accumulated by a numerical amount using the sum.  All
    attributes are queued up in the request_cache for this request.  At the end of
    the request, the monitoring_utils middleware will batch report all
    queued accumulated attributes to the monitoring tool (e.g. New Relic).

    Arguments:
        name (str): The attribute name.  It should be period-delimited, and
            increase in specificity from left to right.  For example:
            'xb_user_state.get_many.num_items'.
        value (number):  The amount to accumulate into the named attribute.  When
            accumulate() is called multiple times for a given attribute name
            during a request, the sum of the values for each call is reported
            for that attribute.  For attributes which don't make sense to accumulate,
            use ``set_custom_attribute`` instead.

    """
    CachedCustomMonitoringMiddleware.accumulate_attribute(name, value)


def increment(name):
    """
    Increment a monitoring custom attribute representing a counter.

    Here we simply accumulate a new custom attribute with a value of 1, and the
    middleware should automatically aggregate this attribute.
    """
    accumulate(name, 1)


def set_custom_attributes_for_course_key(course_key):
    """
    Set monitoring custom attributes related to a course key.

    This is not cached, and only support reporting to New Relic Insights.

    """
    if newrelic:  # pragma: no cover
        newrelic.agent.add_custom_parameter('course_id', str(course_key))
        newrelic.agent.add_custom_parameter('org', str(course_key.org))


def set_custom_attribute(key, value):
    """
    Set monitoring custom attribute.

    This is not cached, and only support reporting to New Relic Insights.

    """
    if newrelic:  # pragma: no cover
        # note: parameter is new relic's older name for attributes
        newrelic.agent.add_custom_parameter(key, value)


def record_exception():
    """
    Records a caught exception to the monitoring system.

    Note: By default, only unhandled exceptions are monitored. This function
    can be called to record exceptions as monitored errors, even if you handle
    the exception gracefully from a user perspective.

    For more details, see:
    https://docs.newrelic.com/docs/agents/python-agent/python-agent-api/recordexception-python-agent-api

    """
    if newrelic:  # pragma: no cover
        newrelic.agent.record_exception()


def background_task(*args, **kwargs):
    """
    Handles monitoring for background tasks that are not passed in through the web server like
    celery and event consuming tasks.

    For more details, see:
    https://docs.newrelic.com/docs/apm/agents/python-agent/supported-features/monitor-non-web-scripts-worker-processes-tasks-functions

    """
    def noop_decorator(func):
        return func

    if newrelic:  # pragma: no cover
        return newrelic.agent.background_task(*args, **kwargs)
    else:
        return noop_decorator



"""
Tests for util.logging
"""

import re
import subprocess

import pytest

from edx_django_utils.logging.internal.log_sensitive import decrypt_log_message, encrypt_for_log, generate_reader_keys


def test_encryption_no_key():
    to_log = encrypt_for_log("Testing testing 1234", None)
    assert to_log == '[encryption failed, no key]'


def test_encryption_round_trip():
    reader_keys = generate_reader_keys()
    reader_public_64 = reader_keys['public']
    reader_private_64 = reader_keys['private']

    to_log = encrypt_for_log("Testing testing 1234", reader_public_64)
    re_base64 = r'[a-zA-Z0-9/+=]'
    assert re.fullmatch(f'\\[encrypted: {re_base64}+\\|{re_base64}+\\]', to_log)

    to_decrypt = to_log.partition('[encrypted: ')[2].rstrip(']')

    decrypted = decrypt_log_message(to_decrypt, reader_private_64)
    assert decrypted == "Testing testing 1234"

    # Also check that decryption still works if someone accidentally
    # copies in the trailing framing "]" character, just as a
    # nice-to-have. (base64 module should handle this already, since
    # it stops reading at the first invalid base64 character.)
    decrypted_again = decrypt_log_message(to_decrypt + ']', reader_private_64)
    assert decrypted_again == "Testing testing 1234"


def test_full_cli(tmp_path):
    def do_call(args, stdin=None):
        return subprocess.run(
            ['log-sensitive', *args], check=True,
            input=(stdin.encode() if stdin else None), capture_output=True,
        ).stdout.decode()

    # Generate keys and save the private key to file
    genkeys_out = do_call(['gen-keys'])
    gen_pub_64 = re.search(r'YOUR_DEBUG_PUBLIC_KEY[^"]*"([^"]*)"', genkeys_out).group(1)
    gen_priv_64 = re.search(r'"([^"]*)" \(private\)', genkeys_out).group(1)
    priv_key_file = tmp_path / 'log_sensitive_private.key'
    with open(priv_key_file, 'w') as priv_f:
        priv_f.write(gen_priv_64)

    sample_plaintext = "The Magic Words are Squeamish Ossifrage"

    encrypted_raw = do_call(
        ['encrypt', '--public-key', gen_pub_64, '--message-file', '-'],
        stdin=sample_plaintext
    )

    # Check that there's a useful error message if the [encrypted: ...] wrapper is left in
    with pytest.raises(subprocess.CalledProcessError) as e:
        decrypted = do_call(
            ['decrypt', '--private-key-file', priv_key_file, '--message-file', '-'],
            stdin=encrypted_raw
        )
    assert 'Only include the Base64' in e.value.stderr.decode()

    # Get rid of [encrypted: ...] wrapper/delimiter
    encrypted = encrypted_raw.split(' ')[1].rstrip(']')
    decrypted = do_call(
        ['decrypt', '--private-key-file', priv_key_file, '--message-file', '-'],
        stdin=encrypted
    )

    # Decrypted output comes out of CLI with an extra \n, so get rid of that first
    assert decrypted.strip('\n') == sample_plaintext



"""
Metrics utilities public api

See README.rst for details.
"""
from .internal.code_owner.middleware import CodeOwnerMonitoringMiddleware
from .internal.code_owner.utils import (
    get_code_owner_from_module,
    set_code_owner_attribute,
    set_code_owner_attribute_from_module
)
from .internal.middleware import (
    CachedCustomMonitoringMiddleware,
    CookieMonitoringMiddleware,
    DeploymentMonitoringMiddleware,
    MonitoringMemoryMiddleware
)
from .internal.transactions import (
    function_trace,
    get_current_transaction,
    ignore_transaction,
    set_monitoring_transaction_name
)
from .internal.utils import (
    accumulate,
    background_task,
    increment,
    record_exception,
    set_custom_attribute,
    set_custom_attributes_for_course_key
)
# "set_custom_metric*" methods are deprecated
from .utils import set_custom_metric, set_custom_metrics_for_course_key



from django.db import models


class GenericModel(models.Model):
    name = models.CharField(max_length=100)
    description = models.TextField()
    created = models.DateField()

    class Meta:
        ordering = ('name',)



"""
Management command `manage_group` is used to idempotently create Django groups
and set their permissions by name.
"""


from django.apps import apps
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.utils.translation import gettext as _


class Command(BaseCommand):
    """Creates the specified group, if it does not exist, and sets its permissions."""
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument('group_name')
        parser.add_argument('--remove', dest='is_remove', action='store_true')
        parser.add_argument('-p', '--permissions', nargs='*', default=[])

    def _handle_remove(self, group_name):
        """
        Remove the group named by the string group_name, if it exists.
        """

        try:
            Group.objects.get(name=group_name).delete()
            self.stderr.write(_('Removed group: "{}"').format(group_name))
        except Group.DoesNotExist:
            self.stderr.write(_('Did not find a group with name "{}" - skipping.').format(group_name))

    @transaction.atomic
    def handle(self, group_name, is_remove, permissions=None, **options):  # pylint: disable=arguments-differ

        if is_remove:
            self._handle_remove(group_name)
            return

        old_permissions = set()
        group, created = Group.objects.get_or_create(name=group_name)

        if created:
            try:
                # Needed for sqlite backend (i.e. in tests) because
                # name.max_length won't be enforced by the db.
                # See also http://www.sqlite.org/faq.html#q9
                group.full_clean()
            except ValidationError as exc:
                # give a more helpful error
                raise CommandError(
                    _(
                        'Invalid group name: "{group_name}". {messages}'
                    ).format(
                        group_name=group_name,
                        messages=exc.messages[0]
                    )
                ) from None
            self.stderr.write(_('Created new group: "{}"').format(group_name))
        else:
            self.stderr.write(_('Found existing group: "{}"').format(group_name))
            old_permissions = set(group.permissions.all())

        new_permissions = self._resolve_permissions(permissions or set())

        add_permissions = new_permissions - old_permissions
        remove_permissions = old_permissions - new_permissions

        self.stderr.write(
            _(
                'Adding {codenames} permissions to group "{group}"'
            ).format(
                codenames=[ap.name for ap in add_permissions],
                group=group.name
            )
        )
        self.stderr.write(
            _(
                'Removing {codenames} permissions from group "{group}"'
            ).format(
                codenames=[rp.codename for rp in remove_permissions],
                group=group.name
            )
        )

        group.permissions.set(new_permissions)

        group.save()

    def _resolve_permissions(self, permissions):
        """
        Given a list of permission strings in the format 'app_label:model:codename', return a list
        of Permission model instances.
        """
        new_permissions = set()
        for permission in permissions:
            try:
                app_label, model_name, codename = permission.split(':')
            except ValueError:
                # give a more helpful error
                raise CommandError(_(
                    'Invalid permission option: "{}". Please specify permissions '
                    'using the format: app_label:model_name:permission_codename.'
                ).format(permission)) from None
            # this will raise a LookupError if it fails.
            try:
                model_class = apps.get_model(app_label, model_name)
            except LookupError as exc:
                raise CommandError(str(exc)) from None

            # Fetch content type for model, including proxy models.
            content_type = ContentType.objects.get_for_model(model_class, for_concrete_model=False)
            try:
                new_permission = Permission.objects.get(
                    content_type=content_type,
                    codename=codename,
                )
            except Permission.DoesNotExist:
                # give a more helpful error
                raise CommandError(
                    _(
                        'Invalid permission codename: "{codename}".  No such permission exists '
                        'for the model {module}.{model_name}.'
                    ).format(
                        codename=codename,
                        module=model_class.__module__,
                        model_name=model_class.__name__,
                    )
                ) from None
            new_permissions.add(new_permission)
        return new_permissions



"""
Deprecated monitoring helpers for backward-compatibility.

IMPORTANT: No new code should be added to this file.
TODO: Remove this file once this code are no longer used.

"""
import warnings

from .internal.utils import accumulate as internal_accumulate
from .internal.utils import increment as internal_increment
from .internal.utils import set_custom_attribute as internal_set_custom_attribute
from .internal.utils import set_custom_attributes_for_course_key as internal_set_custom_attributes_for_course_key


def accumulate(name, value):
    """
    Deprecated method. Use public API instead.
    """
    msg = "Use 'monitoring.accumulate' in place of 'monitoring.utils.accumulate'."
    warnings.warn(msg, DeprecationWarning)
    internal_set_custom_attribute('deprecated_monitoring_utils', f'accumulate[{name}]')
    internal_accumulate(name, value)


def increment(name):
    """
    Deprecated method. Use public API instead.
    """
    msg = "Use 'monitoring.increment' in place of 'monitoring.utils.increment'."
    warnings.warn(msg, DeprecationWarning)
    internal_set_custom_attribute('deprecated_monitoring_utils', f'increment[{name}]')
    internal_increment(name)


def set_custom_attribute(key, value):
    """
    Deprecated method. Use public API instead.
    """
    msg = "Use 'monitoring.set_custom_attribute' in place of 'monitoring.utils.set_custom_attribute'."
    warnings.warn(msg, DeprecationWarning)
    internal_set_custom_attribute('deprecated_monitoring_utils', f'set_custom_attribute[{key}]')
    internal_set_custom_attribute(key, value)


def set_custom_attributes_for_course_key(course_key):
    """
    Deprecated method. Use public API instead.
    """
    msg = "Use 'monitoring.set_custom_attributes_for_course_key' in place of " \
          "'monitoring.utils.set_custom_attributes_for_course_key'."
    warnings.warn(msg, DeprecationWarning)
    internal_set_custom_attribute(
        'deprecated_monitoring_utils',
        'set_custom_attributes_for_course_key[{}]'.format(str(course_key))
    )
    internal_set_custom_attributes_for_course_key(course_key)


def set_custom_metric(key, value):  # pragma: no cover
    """
    Deprecated method to set monitoring custom attribute.
    """
    msg = "Use 'set_custom_attribute' in place of 'set_custom_metric'."
    warnings.warn(msg, DeprecationWarning)
    internal_set_custom_attribute('deprecated_monitoring_utils', f'set_custom_metric[{key}]')
    internal_set_custom_attribute(key, value)


def set_custom_metrics_for_course_key(course_key):  # pragma: no cover
    """
    Deprecated method to set monitoring custom attributes for course key.
    """
    msg = "Use 'set_custom_attributes_for_course_key' in place of 'set_custom_metrics_for_course_key'."
    warnings.warn(msg, DeprecationWarning)
    internal_set_custom_attribute(
        'deprecated_monitoring_utils',
        'set_custom_metrics_for_course_key[{}]'.format(str(course_key))
    )
    internal_set_custom_attributes_for_course_key(course_key)
